/******************************************************************************/
/*                                                                            */
/*        Project: Boona Taxi                                                 */
/*          Title: BSP tree functions                                         */
/*        Version: 0.3 (010327)                                               */
/*                                                                            */
/*        Started: 15/12/2000                                                 */
/*   Last Updated: 27/03/2001                                                 */
/*                                                                            */
/*         Author: Michael Bonnington                                         */
/*                                                                            */
/******************************************************************************/

#include "./bsp.h"



/* Test if a point is behind or in front of a plane ==========================*/
int bsp::classifyPoint(cVector3f *point, Triangle *plane)
{
   float result;

   cVector3f *vector1 = (cVector3f *)&plane->v1;
   cVector3f direction = (*vector1)-(*point);
   result = (direction)*(plane->normal);   //dot product

   if(result > TOLERANCE) return CP_BACK;
   else if(result < -TOLERANCE) return CP_FRONT;
   else return CP_COINCIDENT;
}


/* Test if a polygon is behind or in front of a plane ========================*/
int bsp::classifyPoly(Triangle *poly, Triangle *plane)
{
   int behind     = 0,
       infront    = 0,
       coincident = 0;
   float result;

   result = classifyPoint(poly->v1, plane);
   if(result = CP_BACK) behind++;
   else if(result = CP_FRONT) infront++;
   else
   {
      behind++;
      infront++;
      coincident++;
   }
   result = classifyPoint(poly->v2, plane);
   if(result = CP_BACK) behind++;
   else if(result = CP_FRONT) infront++;
   else
   {
      behind++;
      infront++;
      coincident++;
   }
   result = classifyPoint(poly->v3, plane);
   if(result = CP_BACK) behind++;
   else if(result = CP_FRONT) infront++;
   else
   {
      behind++;
      infront++;
      coincident++;
   }

   if(coincident == 3) return CP_FRONT;   //this would normally be CP_COINCIDENT
   if(behind == 3) return CP_BACK;
   if(infront == 3) return CP_FRONT;
   return CP_SPLIT;
}


/* Select the best splitter ==================================================*/
Triangle * bsp::selectBestSplitter(Triangle *polyList)
{
   unsigned long best_score = 100000;   //set to a very high value to begin with
   Triangle *splitter     = polyList;
   Triangle *currentPoly  = NULL;
   Triangle *selectedPoly = NULL;

   while(splitter != NULL)
   {
      currentPoly = polyList;
      unsigned long score = 0;
      unsigned int splits = 0, backfaces = 0, frontfaces = 0;

      while(currentPoly != NULL)
      {
         if(currentPoly != splitter)
         {
            switch(classifyPoly(currentPoly, splitter))
            {
               case CP_COINCIDENT:
                  break;

               case CP_BACK:
                  backfaces++;
                  break;

               case CP_FRONT:
                  frontfaces++;
                  break;

               case CP_SPLIT:
                  splits++;
                  break;

               default:
                  break;
            }
         }
         currentPoly = currentPoly->nextPtr;
      }//end of while(currentPoly != NULL)

      score = abs(frontfaces-backfaces)+(splits*8);
      if(score < best_score)
      {
         best_score = score;
         selectedPoly = splitter;
      }
      splitter = splitter->nextPtr;
   }//end of while(splitter!=NULL)

   return selectedPoly;
}


/* Split a polygon ===========================================================*/
/*void bsp::splitPolygon(Triangle *poly, Triangle *plane,
                         Triangle *frontSplit, Triangle *backSplit);
{
}*/


/* Compile the BSP tree ======================================================*/
void bsp::buildBSPTree(node *CurrentNode, Triangle *polyList)
{
   Triangle *polyTest    = polyList;
   Triangle *FrontList   = NULL;
   Triangle *BackList    = NULL;
   Triangle *NextPolygon = NULL;
   Triangle *FrontSplit  = NULL;
   Triangle *BackSplit   = NULL;
   //cVector3f vec1, vec2;
   CurrentNode->splitter = selectBestSplitter(polyList);

   while(polyTest!=NULL)
   {
      NextPolygon = polyTest->nextPtr;   //remember because polytest->Next will be altered
      if(polyTest != CurrentNode->splitter)
      {
         switch(classifyPoly(polyTest, CurrentNode->splitter))
         {
            case CP_FRONT:
               polyTest->nextPtr = FrontList;
               FrontList = polyTest;
               break;

            case CP_BACK:
               polyTest->nextPtr = BackList;
               BackList = polyTest;
               break;

            /*case CP_SPLIT:
               FrontSplit=new Triangle;
               BackSplit=new Triangle;
               ZeroMemory(FrontSplit, sizeof(Triangle));
               ZeroMemory(BackSplit, sizeof(Triangle));
               SplitPolygon(polyTest, CurrentNode->Splitter, FrontSplit, BackSplit);
               delete polyTest;
               FrontSplit->Next=FrontList;
               FrontList=FrontSplit;
               BackSplit->Next=BackList;
               BackList=BackSplit;
               break;*/

            default:
               break;
         }
      }//end of

   polyTest = NextPolygon;
   }//end of while loop

   if(FrontList == NULL)
   {
      node *leafNode = new node;
      ZeroMemory(leafNode, sizeof(leafNode));
      leafNode->isLeaf  = true;
      leafNode->isSolid = false;
      CurrentNode->front = leafNode;
   }
   else
   {
      node *newNode = new node;
      ZeroMemory(newNode, sizeof(newNode));
      newNode->isLeaf = false;
      CurrentNode->front = newNode;
      buildBSPTree(newNode, FrontList);
   }
   if(BackList == NULL)
   {
      node *leafNode = new node;
      ZeroMemory(leafNode, sizeof(leafNode));
      leafNode->isLeaf  = true;
      leafNode->isSolid = true;
      CurrentNode->back = leafNode;
   }
   else
   {
      node *newNode = new node;
      ZeroMemory(newNode, sizeof(newNode));
      newNode->isLeaf = false;
      CurrentNode->back = newNode;
      buildBSPTree(newNode, BackList);
   }
}


/* Walk through the tree and render polygons =================================*/
void bsp::walkBSPTree(node *Node, cVector3f *cam_pos)
{
   if(Node->isLeaf) return;

   int result = classifyPoint(cam_pos, Node->splitter);
   if(result == CP_FRONT)   //change to CP_BACK to reverse rendering order
   {
      if(Node->back != NULL walkBSPTree(Node->back, cam_pos));
      //drawPoly(&Node->splitter, m_pVertexArray, m_pNormalArray, m_pTexCoordArray);
      if(Node->front != NULL walkBSPTree(Node->front, cam_pos));
   }
   else
   {
      if(Node->front != NULL walkBSPTree(Node->front, cam_pos));
      //drawPoly(&Node->splitter, m_pVertexArray, m_pNormalArray, m_pTexCoordArray);
      if(Node->back != NULL walkBSPTree(Node->back, cam_pos));
   }
   return;
}
