/******************************************************************************/
/*                                                                            */
/*        Project: Boona Taxi                                                 */
/*          Title: Game Engine (GLUT callbacks)                               */
/*        Version: 0.3 (010328)                                               */
/*                                                                            */
/*        Started: 19/09/2000                                                 */
/*   Last Updated: 28/03/2001                                                 */
/*                                                                            */
/*         Author: Michael Bonnington                                         */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>

#include "./opengl_includes.h"
#include "./font.h"
#include "./frametime.h"
#include "./callbacks.h"
#include "./car.h"
#include "./tga.h"
#include "./objloader.h"

#include "./fruity_exception.h"
#include "./fruity_math.h"



/* Constants =================================================================*/

#define FARPLANE 256   //the farthest viewing distance

/* Constants for game modes (add/remove/change as required)... */
#define INTRO 8
#define TITLES 5
#define SELECT 0
#define OPTIONS 1
#define HIGHSCORES 2
#define CREDITS 3
#define EXIT 4
#define GAME 6
#define PAUSED 7

/* Constants for camera views... */
#define IN_CAR 0
#define CHASE 1
#define COPTER 2

//get rid of these when car class is sorted...
#define ETYPE 0
#define ESPRIT 1
#define PUG 2
#define BUG 3



/* Macros ====================================================================*/
#define resetView(a) (a).spin_x  = 0;\
                     (a).spin_y  = 0;\
                     (a).dolly   = 0;\
                     (a).trans_x = 0;\
                     (a).trans_y = 0;



/* Function prototypes =======================================================*/
void changeGameMode(int);

/* Drawing functions... */
void drawCity(void);
void drawOverlay(void);
void drawHUD(void);

/* Game mode-specific GLUT callback functions... */
void idleIntro(void);
void displayIntro(void);

void idleTitles(void);
void displayTitles(void);

void idleSelect(void);
void displaySelect(void);

void idleOptions(void);
void displayOptions(void);

void idleHighscores(void);
void displayHighscores(void);

void idleCredits(void);
void displayCredits(void);

void idleGame(void);
void displayGame(void);

void idlePaused(void);
void displayPaused(void);

/* General GLUT callback functions... */
void reshape(int, int);
void keyDown(unsigned char, int, int);
void keyUp(unsigned char, int, int);
void specialDown(int, int, int);
void specialUp(int, int, int);
void mouse(int, int, int, int);
void motion(int, int);



/* Globals ===================================================================*/

/* Structure to hold window information, including size and position... */
struct window win;

/* Display list numbers to hold perspective and orthographic view set-up... */
GLuint persp, ortho;

/* Display list numbers for textures not associated with models... */
GLuint backgroundTexture, logoTexture, groundTexture;// envMapTexture;

/* Game class (defined in callbacks.h header file)... */
game boona;


bool keys[128], specials[128];
char statusLine[256];      //line of text to display messages on-screen
int current_car = ETYPE;   //remove car when I have car class


bool level_map[98][166];   //blag - used for collisions
objFile city, pavement;
car player(0, 0, 0);
camera cam;



/* Game class initialisation =================================================*/
void game::initData(void)
{
   /* Load data from config files... */
   if(!loadData())
   {
      /* Set default values... */
      lighting_s     = true;
      lighting_d     = true;
      use_env_map    = true;
      use_fog        = true;
      fog_colour[0]  = 0.46667;
      fog_colour[1]  = 0.73333;
      fog_colour[2]  = 1.0;
      fog_colour[3]  = 1.0;
      texture_filter = 1;
      time_limit     = 300;
      show_hud       = true;
      show_fps       = true;
      cam_view       = CHASE;
   }

   boona.rank = 9; //temp
}

/* Load config data ==========================================================*/
bool game::loadData(void)
{
   ifstream level_data,
            render_opt,
            game_opt,
            highscores;
   int x, z;
   char in_str[80];
   bool success = true;


   /* Read in "RenderingOptions.cfg" file... */
   render_opt.open("./config/RenderingOptions.cfg", ios::in);

   if(render_opt.fail()) success = false;
   else   //load values from config file
   {
      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) lighting_s = true;
      else lighting_s = false;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) lighting_d = true;
      else lighting_d = false;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) use_env_map = true;
      else use_env_map = false;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) use_fog = true;
      else use_fog = false;

      render_opt >> in_str
                 >> fog_colour[0]
                 >> fog_colour[1]
                 >> fog_colour[2]
                 >> fog_colour[3];

      render_opt >> in_str >> texture_filter;

      render_opt.close();
   }


   /* Read in "GameOptions.cfg" file... */
   game_opt.open("./config/GameOptions.cfg", ios::in);

   if(game_opt.fail()) success = false;
   else   //load values from config file
   {
      game_opt >> in_str >> time_limit;

      game_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) show_hud = 1;
      else show_hud = 0;

      game_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) show_fps = 1;
      else show_fps = 0;

      game_opt >> in_str >> cam_view;

      game_opt.close();
   }


   /* Read in "BoonaHighscores.cfg" file... */
   highscores.open("./config/BoonaHighscores.cfg", ios::in);

   if(highscores.fail()) success = false;
   else   //load values from config file
   {
      highscores >> in_str
                 >> highscore[0].name >> highscore[0].car >> highscore[0].score;
      highscores >> in_str
                 >> highscore[1].name >> highscore[1].car >> highscore[1].score;
      highscores >> in_str
                 >> highscore[2].name >> highscore[2].car >> highscore[2].score;
      highscores >> in_str
                 >> highscore[3].name >> highscore[3].car >> highscore[3].score;
      highscores >> in_str
                 >> highscore[4].name >> highscore[4].car >> highscore[4].score;
      highscores >> in_str
                 >> highscore[5].name >> highscore[5].car >> highscore[5].score;
      highscores >> in_str
                 >> highscore[6].name >> highscore[6].car >> highscore[6].score;
      highscores >> in_str
                 >> highscore[7].name >> highscore[7].car >> highscore[7].score;
      highscores >> in_str
                 >> highscore[8].name >> highscore[8].car >> highscore[8].score;
      highscores >> in_str
                 >> highscore[9].name >> highscore[9].car >> highscore[9].score;

      highscores.close();
   }


   /* Read in the level data from a text file ("levelmap.txt")... */
   level_data.open("./config/levelmap.txt", ios::in);

   if(level_data.fail()) success = false;
   else   //load values from text file
   {
      for(z=0; z<166; z++)
      {
         for(x=0; x<98; x++)
         {
            level_data >> level_map[x][z];
         }
      }
      level_data.close();
   }

   return success;
}

/* Save config data ==========================================================*/
bool game::saveData(void)
{
   ofstream highscores;
   bool success = true;


   /* Save out "BoonaHighscores.cfg" file... */
   highscores.open("./config/BoonaHighscores1.cfg", ios::out);

   if(highscores.fail()) success = false;
   else   //save values to config file
   {
      for(int i=0; i<10; i++)
         highscores << i+1 << " "
                    << highscore[i].name << " "
                    << highscore[0].car << " "
                    << highscore[0].score << endl;

      /*highscores >> in_str
                 >> highscore[0].name >> highscore[0].car >> highscore[0].score;
      highscores >> in_str
                 >> highscore[1].name >> highscore[1].car >> highscore[1].score;
      highscores >> in_str
                 >> highscore[2].name >> highscore[2].car >> highscore[2].score;
      highscores >> in_str
                 >> highscore[3].name >> highscore[3].car >> highscore[3].score;
      highscores >> in_str
                 >> highscore[4].name >> highscore[4].car >> highscore[4].score;
      highscores >> in_str
                 >> highscore[5].name >> highscore[5].car >> highscore[5].score;
      highscores >> in_str
                 >> highscore[6].name >> highscore[6].car >> highscore[6].score;
      highscores >> in_str
                 >> highscore[7].name >> highscore[7].car >> highscore[7].score;
      highscores >> in_str
                 >> highscore[8].name >> highscore[8].car >> highscore[8].score;
      highscores >> in_str
                 >> highscore[9].name >> highscore[9].car >> highscore[9].score;
*/
      highscores.close();
   }


   return success;
}



/* GLUT callback registration ================================================*/
void registerCallbacks(void)
{
   InitErrorLog();

   printf("Initialising OpenGL:\n  Vendor: %s\nRenderer: %s\n Version: %s\n\n",
          (char *)glGetString(GL_VENDOR),
          (char *)glGetString(GL_RENDERER),
          (char *)glGetString(GL_VERSION));


   /* Initialise OpenGL/GLUT stuff... */

   /* Enable backface culling... */
   glEnable(GL_CULL_FACE);
   glCullFace(GL_BACK);

   /* Lighting setup... */
   float pAmbient[]  = {0.3, 0.3,  0.3, 1.0},
         pDiffuse[]  = {0.7, 0.7,  0.7, 1.0},
         pSpecular[] = {1.0, 1.0,  1.0, 1.0},
         pPosition[] = {0.0, 16.0, 8.0, 1.0};

   glLightfv(GL_LIGHT0, GL_AMBIENT,  pAmbient);
   glLightfv(GL_LIGHT0, GL_AMBIENT,  pDiffuse);
   glLightfv(GL_LIGHT0, GL_AMBIENT,  pSpecular);
   glLightfv(GL_LIGHT0, GL_POSITION, pPosition);

   glMaterialfv(GL_FRONT, GL_AMBIENT,   pAmbient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE,   pDiffuse);
   glMaterialfv(GL_FRONT, GL_SPECULAR,  pSpecular);
   glMaterialf (GL_FRONT, GL_SHININESS, 5);

   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

   /* Set up automatic texture generation for environment mapping... */
   glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
   glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
   glDepthFunc(GL_LEQUAL);

   /* Fog setup... */
   glClearColor(boona.fog_colour[0],
                boona.fog_colour[1],
                boona.fog_colour[2],
                boona.fog_colour[3]);
   glFogi(GL_FOG_MODE, GL_LINEAR);
   glFogfv(GL_FOG_COLOR, boona.fog_colour);
   glFogf(GL_FOG_DENSITY, 0.05);
   glFogf(GL_FOG_END, FARPLANE);

   glutIgnoreKeyRepeat(1);
   glutSetCursor(GLUT_CURSOR_NONE);


   /* Initialise frame counter... */
   DebugLog("Initialising frame counter...");
   if(!initFrameCounter())
   {
      printf("ERROR: Unable to initialise frame counter.\n\n");
      changeGameMode(EXIT);
   }

   DebugLog("Making fonts...");
   makeRasterFont();   //initialise font

   /*DebugLog("Initialising renderer...");
   initRenderer();   - POINTLESS! */

   /* Load the models... */
   DebugLog("Loading models...");
   if(!city.load("./models/city.obj"))
      printf("WARNING: An error occured while reading city.obj.\n\n");
   if(!city.LoadMD("./models/city.mdf"))
      printf("WARNING: An error occured while reading city.mdf.\n\n");
   if(!pavement.load("./models/pavement.obj"))
      printf("WARNING: An error occured while reading pavement.obj.\n\n");
   if(!pavement.LoadMD("./models/pavement.mdf"))
      printf("WARNING: An error occured while reading pavement.mdf.\n\n");
   if(!player.load("./models/etype.obj"))
      printf("WARNING: An error occured while reading etype.obj.\n\n");
   if(!player.LoadMD("./models/etype.mdf"))
      printf("WARNING: An error occured while reading etype.mdf.\n\n");

   /* Load textures not associated with models... */
   DebugLog("Loading textures...");
   loadTexture("./textures/background_logo.tga", &backgroundTexture);
   loadTexture("./textures/boonataxi_logo.tga", &logoTexture);
   loadTexture("./textures/ground1.tga", &groundTexture);

   boona.option = SELECT;


   /* Register GLUT callback functions... */
   DebugLog("Registering GLUT callback functions...");
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyDown);
   glutKeyboardUpFunc(keyUp);
   glutSpecialFunc(specialDown);
   glutSpecialUpFunc(specialUp);
   changeGameMode(INTRO);

   glutPostRedisplay();
}



/* Function to change current game mode ======================================*/
void changeGameMode(int mode)
{
   switch(mode)
   {
      case INTRO:
         boona.game_mode     = INTRO;
         boona.msg_counter   = 8;             //seconds to display intro for
         boona.scroll_offset = win.width*4;   //initial size of logo
         player.reset(-12, 0, -48, 2);
         resetView(boona);
         glutIdleFunc(idleIntro);
         glutDisplayFunc(displayIntro);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case TITLES:
         boona.game_mode     = TITLES;
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
         resetView(boona);
         glutIdleFunc(idleTitles);
         glutDisplayFunc(displayTitles);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case SELECT:
         boona.game_mode = SELECT;
         resetView(boona);
         glutIdleFunc(idleSelect);
         glutDisplayFunc(displaySelect);
         glutMouseFunc(mouse);
         glutMotionFunc(motion);
         break;

      case OPTIONS:
         boona.game_mode     = OPTIONS;
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
         resetView(boona);
         glutIdleFunc(idleOptions);
         glutDisplayFunc(displayOptions);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case HIGHSCORES:
         boona.game_mode     = HIGHSCORES;
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
         resetView(boona);
         glutIdleFunc(idleHighscores);
         glutDisplayFunc(displayHighscores);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case CREDITS:
         boona.game_mode     = CREDITS;
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
         resetView(boona);
         glutIdleFunc(idleCredits);
         glutDisplayFunc(displayCredits);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case EXIT:   //use "changeGameMode(EXIT)" instead of "exit(0)" to quit...
         boona.game_mode = EXIT;
         /* Save highscores table... */
         if(!boona.saveData())
            printf("WARNING: Unable to save highscores.\n\n");
         //seems to crash on exit when using game mode under Windows 2000...
         if(glutGameModeGet(GLUT_GAME_MODE_ACTIVE)) glutLeaveGameMode();
         exit(0);
         break;

      case GAME:
         boona.game_mode = GAME;
         resetView(boona);
         glutIdleFunc(idleGame);
         glutDisplayFunc(displayGame);
         glutMouseFunc(NULL);
         glutMotionFunc(NULL);
         break;

      case PAUSED:
         boona.game_mode = PAUSED;
         glutIdleFunc(idlePaused);
         glutDisplayFunc(displayPaused);
         glutMouseFunc(mouse);
         glutMotionFunc(motion);
         break;

      default:
         break;
   }
}



/* Function to draw scene (level, cars, etc) =================================*/
void drawScene(void)
{
   glPushMatrix();

    //place camera at the correct position...
    if(boona.top_down_view)
       gluLookAt(player.pos_x, player.pos_y+40, player.pos_z,
                 player.pos_x, player.pos_y   , player.pos_z,
                 1, 0, 0);
    else
       gluLookAt(cam.pos.x, cam.pos.y, cam.pos.z,
                 cam.aim.x, cam.aim.y, cam.aim.z,
                 0, 1, 0);

    //draw city, other cars, etc...
    if(boona.use_fog) glEnable(GL_FOG);
    city.draw();
    pavement.draw();
      //drawOtherStuffHere();
    player.drawCar(boona.lighting_d);
      //drawOtherCarsHere();
    if(boona.use_fog) glDisable(GL_FOG);

    //draw big quad on the ground so we can see what's going on (temporary)...
    glEnable(GL_TEXTURE_2D);
    glColor3f(0.7, 0.7, 0.7);
    glBindTexture(GL_TEXTURE_2D, groundTexture);
    glBegin(GL_QUADS);
     glTexCoord2f(0, 166);
     glVertex3f(0, -0.3, 0);
     glTexCoord2f(0, 0);
     glVertex3f(0, -0.3, 166);
     glTexCoord2f(98, 0);
     glVertex3f(98, -0.3, 166);
     glTexCoord2f(98, 166);
     glVertex3f(98, -0.3, 0);
    glEnd();
    glDisable(GL_TEXTURE_2D);

    //draw grid squares for collisions (temporary)...
    /*glEnable(GL_TEXTURE_2D);
    glColor3f(0.7, 0.7, 0.7);
    glBindTexture(GL_TEXTURE_2D, groundTexture);
    for(int z=0; z<166; z++)
    {
       for(int x=0; x<98; x++)
       {
          if(!level_map[x][z])
          {
             glBegin(GL_QUADS);
              glTexCoord2f(0, 1);
              glVertex3f(x+0, -0.01, z+0);
              glTexCoord2f(0, 0);
              glVertex3f(x+0, -0.01, z+1);
              glTexCoord2f(1, 0);
              glVertex3f(x+1, -0.01, z+1);
              glTexCoord2f(1, 1);
              glVertex3f(x+1, -0.01, z+0);
             glEnd();
          }
       }
    }
    glDisable(GL_TEXTURE_2D);*/

   glPopMatrix();
}

/* Function to draw presentation screen overlay graphics =====================*/
void drawOverlay(void)
{
   /* Make sure the orthographic display list has been initialised with
    * 'glCallList(persp);'
    * before trying to call this function.
    */

   glBegin(GL_QUADS);
    glColor3f(0.93333, 0.66667, 0.0);
    glVertex2f(0, win.height-44);
    glVertex2f(0, win.height);
    glVertex2f(win.width, win.height);
    glVertex2f(win.width, win.height-44);

    glColor3f(1.0, 0.8, 0.06667);
    glVertex2f(0, win.height-128);
    glVertex2f(0, win.height-44);
    glVertex2f(win.width, win.height-44);
    glVertex2f(win.width, win.height-128);

    glVertex2f(0, 0);
    glVertex2f(0, 64);
    glVertex2f(win.width, 64);
    glVertex2f(win.width, 0);
   glEnd();

   smallFont("www.boonaracer.co.uk/taxi", 16, win.height-100, 0.0, 0.0, 0.0);
   smallFont(statusLine, 16, win.height-16, 0.0, 0.0, 0.0);
}

/* Function to draw heads up display =========================================*/
void drawHUD(void)
{
   /* Make sure the orthographic display list has been initialised with
    * 'glCallList(persp);'
    * before trying to call this function.
    */

   //try putting some of this in display lists to speed it up...

   char info_text[8];

   sprintf(info_text, "$%4.2f", player.getScore());
   smallFont("Total Earned", 16, 28, 1.0, 1.0, 1.0);
   largeOutlineFont(info_text, 16, 60, 1.0, 0.8, 0.06667);

   sprintf(info_text, "$%4.2f", player.getScore());   //change to fare
   smallFont("This Fare", 16, 76, 1.0, 1.0, 1.0);
   largeOutlineFont(info_text, 16, 108, 1.0, 0.6, 0.0);

   sprintf(info_text, "%d", boona.rank);
   smallFont("Rank", 16, 124, 1.0, 1.0, 1.0);
   largeOutlineFont(info_text, 16, 156, 1.0, 0.8, 0.06667);

   sprintf(info_text, "%d", (int)boona.game_time);
   smallFont("Game Time", win.width-96, 28, 1.0, 1.0, 1.0);
   largeOutlineFont(info_text, win.width-96, 60, 1.0, 0.8, 0.06667);

   //draw current time in centre of screen

   sprintf(info_text, "%d", (int)(player.getSpeed()*10));
   smallFont("Speed", win.width-96, win.height-48, 1.0, 1.0, 1.0);
   largeOutlineFont(info_text, win.width-96, win.height-16, 1.0, 0.8, 0.06667);

   if(player.getReverse())   //Are we in reverse?
      largeOutlineFont("R", win.width-32, win.height-16, 1.0, 0.6, 0.0);
   else                 //Are we in drive?
      largeOutlineFont("D", win.width-32, win.height-16, 0.33333, 0.8, 0.33333);

   //draw rotating map here
}

/* Interpolate camera position ===============================================
void interpolateCam(float new_pos_x, float new_pos_y, float new_pos_z,
                    float new_aim_x, float new_aim_y, float new_aim_z,
                    float amount)
{
   cam.pos.x = cam.old_pos.x*(1-amount) + new_pos_x*amount;
   cam.pos.y = cam.old_pos.y*(1-amount) + new_pos_y*amount;
   cam.pos.z = cam.old_pos.z*(1-amount) + new_pos_z*amount;
   cam.pos.x = cam.old_aim.x*(1-amount) + new_aim_x*amount;
   cam.pos.y = cam.old_aim.y*(1-amount) + new_aim_y*amount;
   cam.pos.z = cam.old_aim.z*(1-amount) + new_aim_z*amount;
}*/

/* Update and interpolate camera position ====================================*/
void updateCam(void)
{
   /*float new_pos_x, new_pos_y, new_pos_z, new_aim_x, new_aim_y, new_aim_z;*/

   cVector3f new_pos, new_aim;

   //not needed at present...
   /* cam.distance = sqrt((cam.aim_x - cam.pos_x)*(cam.aim_x - cam.pos_x) +
                          (cam.aim_z - cam.pos_z)*(cam.aim_z - cam.pos_z)); */

   //these values need tweaking...
   switch(boona.cam_view)
   {
      case IN_CAR:
         /*new_pos_x = player.pos_x;
         new_pos_y = player.pos_y + 0.85;
         new_pos_z = player.pos_z;
         new_aim_x = player.pos_x + player.dx;
         new_aim_y = player.pos_y + 0.85;
         new_aim_z = player.pos_z + player.dz;*/
         new_pos.set(player.pos_x,
                     player.pos_y + 0.85,
                     player.pos_z);
         new_aim.set(player.pos_x + player.dx,
                     player.pos_y + 0.85,
                     player.pos_z + player.dz);
         break;

      case CHASE:
         /*new_pos_x = player.pos_x - player.dx*4;
         new_pos_y = player.pos_y + 1.6;
         new_pos_z = player.pos_z - player.dz*4;
         new_aim_x = player.pos_x;
         new_aim_y = player.pos_y + 1.2;
         new_aim_z = player.pos_z;*/
         new_pos.set(player.pos_x - player.dx*4,
                     player.pos_y + 1.6,
                     player.pos_z - player.dz*4);
         new_aim.set(player.pos_x,
                     player.pos_y + 1.2,
                     player.pos_z);
         break;

      case COPTER:
         /*new_pos_x = player.pos_x - player.dx*4;
         new_pos_y = player.pos_y + 4;
         new_pos_z = player.pos_z - player.dz*4;
         new_aim_x = player.pos_x;
         new_aim_y = player.pos_y + 2;
         new_aim_z = player.pos_z;*/
         new_pos.set(player.pos_x - player.dx*4,
                     player.pos_y + 4,
                     player.pos_z - player.dz*4);
         new_aim.set(player.pos_x,
                     player.pos_y + 2,
                     player.pos_z);
         break;

      default:
         break;
   }

/* The next six lines take the new position and aiming coordinates for the
 * camera and smoothly interpolates from its current position to the new
 * position over a period of time. Only linear interpolation at the moment -
 * could be re-written to use quadratic or cubic interpolation for a smoother
 * effect...
 */

   /*cam.pos.x = cam.old_pos.x*(1-boona.cam_counter)
             + new_pos_x*boona.cam_counter;
   cam.pos.y = cam.old_pos.y*(1-boona.cam_counter)
             + new_pos_y*boona.cam_counter;
   cam.pos.z = cam.old_pos.z*(1-boona.cam_counter)
             + new_pos_z*boona.cam_counter;
   cam.aim.x = cam.old_aim.x*(1-boona.cam_counter)
             + new_aim_x*boona.cam_counter;
   cam.aim.y = cam.old_aim.y*(1-boona.cam_counter)
             + new_aim_y*boona.cam_counter;
   cam.aim.z = cam.old_aim.z*(1-boona.cam_counter)
             + new_aim_z*boona.cam_counter;*/

   cam.pos = cam.old_pos*(1-boona.cam_counter)
             + new_pos*boona.cam_counter;
   cam.aim = cam.old_aim*(1-boona.cam_counter)
             + new_aim*boona.cam_counter;
}

/* ===========================================================================*/
void positionCam(void)
{
   switch(boona.cam_view)
   {
      case IN_CAR:
         cam.pos.x = player.pos_x;
         cam.pos.y = player.pos_y + 0.85;
         cam.pos.z = player.pos_z;
         cam.aim.x = player.pos_x + player.dx;
         cam.aim.y = player.pos_y + 0.85;
         cam.aim.z = player.pos_z + player.dz;
         break;

      case CHASE:
         cam.pos.x = player.pos_x - player.dx*4;
         cam.pos.y = player.pos_y + 1.6;
         cam.pos.z = player.pos_z - player.dz*4;
         cam.aim.x = player.pos_x;
         cam.aim.y = player.pos_y + 1.2;
         cam.aim.z = player.pos_z;
         break;

      case COPTER:
         cam.pos.x = player.pos_x - player.dx*4;
         cam.pos.y = player.pos_y + 4;
         cam.pos.z = player.pos_z - player.dz*4;
         cam.aim.x = player.pos_x;
         cam.aim.y = player.pos_y + 2;
         cam.aim.z = player.pos_z;
         break;

      default:
         break;
   }
}



/*********************** ******************************************************/
/* INTRO               * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Intro screen) ==============================================*/
void idleIntro(void)
{
   double frametime = getFrametime();

   if(keys[13] || keys[27])
   {
      changeGameMode(TITLES);
      keys[13] = false;
      keys[27] = false;
   }

   player.setAccelerator(true);
   player.update(frametime/2.4);

   //updateCam();
   //cam.pos.z += frametime;
         cam.pos.x = player.pos_x;
         cam.pos.y = player.pos_y + 0.2 + (boona.msg_counter/8);
         cam.pos.z = boona.msg_counter*6;
         cam.aim.x = player.pos_x;
         cam.aim.y = player.pos_y + 1;
         cam.aim.z = player.pos_z;

   if(boona.msg_counter>0)
      boona.msg_counter -= frametime;
   else
   {
      boona.msg_counter = 0;

      player.reset(-12, 0, 47.5, 0);
      changeGameMode(TITLES);
      cam.pos.x = player.pos_x;
      cam.pos.y = player.pos_y + 1.2;
      cam.pos.z = 48;
      cam.aim.x = player.pos_x;
      cam.aim.y = player.pos_y + 1;
      cam.aim.z = player.pos_z;
   }

   if(boona.msg_counter<2)
   {
      if(boona.scroll_offset > 0)
         boona.scroll_offset -= frametime*(win.width*2);
      else
         boona.scroll_offset = 0;
   }

   glutPostRedisplay();
}

/* Display function (Intro screen) ===========================================*/
void displayIntro(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);

   glPushMatrix();
    drawScene();
   glPopMatrix();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   /* Draw big Boona Taxi title logo... */
   if(boona.msg_counter<2)
   {
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glEnable(GL_BLEND);
      glEnable(GL_TEXTURE_2D);
      glColor3f(1.0, 1.0, 1.0);
      glBindTexture(GL_TEXTURE_2D, logoTexture);
      glBegin(GL_QUADS);
       glTexCoord2f(0, 1);
       glVertex2f(128-boona.scroll_offset,
                  64-boona.scroll_offset);
       glTexCoord2f(0, 0);
       glVertex2f(128-boona.scroll_offset,
                  win.height-128+boona.scroll_offset);
       glTexCoord2f(1, 0);
       glVertex2f(win.width-128+boona.scroll_offset,
                  win.height-128+boona.scroll_offset);
       glTexCoord2f(1, 1);
       glVertex2f(win.width-128+boona.scroll_offset,
                  64-boona.scroll_offset);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);
   }

   largeOutlineFont("INSERT COIN", (win.width/2)-88, win.height-32,
                    1.0, 1.0, 1.0);

   glutSwapBuffers();
}

/******************************************************************************/
/* end of INTRO                                                               */
/******************************************************************************/



/*********************** ******************************************************/
/* TITLES              * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Title screen) ==============================================*/
void idleTitles(void)
{
   double frametime = getFrametime();

   if(specials[GLUT_KEY_LEFT])
   {
      if(boona.option > SELECT) boona.option--;
      else boona.option = EXIT;
      boona.msg_counter = 0;
      specials[GLUT_KEY_LEFT] = false;
   }

   if(specials[GLUT_KEY_RIGHT])
   {
      if(boona.option < EXIT) boona.option++;
      else boona.option = SELECT;
      boona.msg_counter = 0;
      specials[GLUT_KEY_RIGHT] = false;
   }

   if(keys[13])
   {
      changeGameMode(boona.option);
      boona.msg_counter = 0;
      keys[13] = false;
   }

   if(keys[27])
   {
      if(boona.option == EXIT) changeGameMode(EXIT);
      else boona.option = EXIT;   //don't exit immediately
      boona.msg_counter = 0;
      keys[27] = false;
   }

   if(boona.msg_counter<8)
      boona.msg_counter+=frametime;
   else
      changeGameMode(INTRO);

   glutPostRedisplay();
}

/* Display function (Title screen) ===========================================*/
void displayTitles(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);
   drawScene();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   /* Draw big Boona Taxi title logo... */
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glEnable(GL_BLEND);
   glEnable(GL_TEXTURE_2D);
   glColor3f(1.0, 1.0, 1.0);
   glBindTexture(GL_TEXTURE_2D, logoTexture);
   glBegin(GL_QUADS);
    glTexCoord2f(0, 1);
    glVertex2f(128, 64);
    glTexCoord2f(0, 0);
    glVertex2f(128, win.height-128);
    glTexCoord2f(1, 0);
    glVertex2f(win.width-128, win.height-128);
    glTexCoord2f(1, 1);
    glVertex2f(win.width-128, 64);
   glEnd();
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_BLEND);

   drawOverlay();
   largeFont("Boona Taxi Title Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeFont("<          >", 0, win.height-60, 0.93333, 0.66667, 0.0);

   switch(boona.option)
   {
      case SELECT:
         largeOutlineFont("Begin Game", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Press 'ENTER' to begin a new game; 'LEFT' or 'RIGHT' to select another option");
         break;

      case OPTIONS:
         largeOutlineFont("Options", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Modify the game settings");
         break;

      case HIGHSCORES:
         largeOutlineFont("Highscores", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "See the ten best Boona Taxi drivers");
         break;

      case CREDITS:
         largeOutlineFont("Credits", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "People who worked on Boona Taxi");
         break;

      case EXIT:
         largeOutlineFont("Exit Game", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Press 'ENTER' or 'ESC' to exit Boona Taxi");
         break;

      default:
         break;
   }

   glutSwapBuffers();
}

/******************************************************************************/
/* end of TITLES                                                              */
/******************************************************************************/



/*********************** ******************************************************/
/* SELECT              * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Car select screen) =========================================*/
void idleSelect(void)
{
   double frametime = getFrametime();

   if(specials[GLUT_KEY_LEFT])
   {
      if(current_car > ETYPE) current_car--;
      else current_car = BUG;
      specials[GLUT_KEY_LEFT] = false;
   }

   if(specials[GLUT_KEY_RIGHT])
   {
      if(current_car < BUG) current_car++;
      else current_car = ETYPE;
      specials[GLUT_KEY_RIGHT] = false;
   }

   if(keys[13])   //enter - select car and proceed to game
   {
      //load car model(s) and data...
      switch(current_car)
      {
         case(ETYPE):
            /*if(!player.load("./models/etype.obj"))
               printf("WARNING: An error occured while reading etype.obj.\n\n");
            if(!player.LoadMD("./models/etype.mdf"))
               printf("WARNING: An error occured while reading etype.mdf.\n\n");*/
            break;

         case(ESPRIT):
            /*if(!player.load("./models/esprit.obj"))
               printf("WARNING: An error occured while reading esprit.obj.\n\n");
            if(!player.LoadMD("./models/esprit.mdf"))
               printf("WARNING: An error occured while reading esprit.mdf.\n\n");*/
            break;

         case(PUG):
            /*if(!player.load("./models/etype.obj"))
               printf("WARNING: An error occured while reading etype.obj.\n\n");
            if(!player.LoadMD("./models/etype.mdf"))
               printf("WARNING: An error occured while reading etype.mdf.\n\n");*/
            break;

         case(BUG):
            /*if(!player.load("./models/etype.obj"))
               printf("WARNING: An error occured while reading etype.obj.\n\n");
            if(!player.LoadMD("./models/etype.mdf"))
               printf("WARNING: An error occured while reading etype.mdf.\n\n");*/
            break;

      }

      boona.game_time = boona.time_limit;
      player.reset(16, 0, 16, 0);
      changeGameMode(GAME);

      keys[13] = false;
   }

   if(keys[27])
   {
      changeGameMode(TITLES);
      keys[27] = false;
   }

   /* Calculate scrolling background position... */
   if(boona.cam_counter < 1-frametime/4) boona.cam_counter+=frametime/4;
   else boona.cam_counter = 0;

   /* Calculate car rotation... */
   boona.spin_y += frametime*16;

   glutPostRedisplay();
}

/* Display function (Car select screen) ======================================*/
void displaySelect(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw scrolling background... */
   glCallList(ortho);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, backgroundTexture);
   glBegin(GL_QUADS);
   glTexCoord2f(0+boona.cam_counter, 1.5+boona.cam_counter);
   glVertex2f(0, 64);
   glTexCoord2f(0+boona.cam_counter, 0+boona.cam_counter);
   glVertex2f(0, win.height-128);
   glTexCoord2f(2.5+boona.cam_counter, 0+boona.cam_counter);
   glVertex2f(win.width, win.height-128);
   glTexCoord2f(2.5+boona.cam_counter, 1.5+boona.cam_counter);
   glVertex2f(win.width, 64);
   glEnd();

   /* Draw all the 3D stuff... */
   glCallList(persp);

   glPushMatrix();
    //glColor3f(1.0, 1.0, 1.0);
    gluLookAt(0, 0.8, 2.6, 0, 0, 0, 0, 1, 0);

    glPushMatrix();
     glTranslatef(boona.trans_x, boona.trans_y, boona.dolly);
     glRotatef(boona.spin_x, 1.0, 0.0, 0.0);

     glPushMatrix();
      glRotatef(boona.spin_y, 0.0, 1.0, 0.0);
      if(boona.lighting_d)
      {
         glEnable(GL_LIGHTING);
         glEnable(GL_LIGHT0);
      }
      player.draw();   //Draw the car model
      if(boona.lighting_d)
      {
         glDisable(GL_LIGHTING);
         glDisable(GL_LIGHT0);
      }
     glPopMatrix();

    glPopMatrix();

   glPopMatrix();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   drawOverlay();
   largeFont("Boona Taxi Vehicle Select Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeFont("<            >", 0, win.height-60, 0.93333, 0.66667, 0.0);

   //change this when I have a proper car class sorted out...
   switch(current_car)
   {
      case ETYPE:
         //player.LoadModel("./models/etype.obj");
         largeOutlineFont("E-type Jag", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "E-type Jag");
         break;

      case ESPRIT:
         //player.LoadModel("./models/esprit.obj");
         largeOutlineFont("Lotus Esprit", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Lotus Esprit");
         break;

      case PUG:
         //player.LoadModel("./models/peugeot.obj");
         largeOutlineFont("Peugeot 406", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Peugeot 406");
         break;

      case BUG:
         //player.LoadModel("./models/bug.obj");
         largeOutlineFont("Bad Boy Bug", 16, win.height-60, 1.0, 1.0, 1.0);
         sprintf(statusLine, "Bad Boy Bug");
         break;

      default:
         break;
   }

   glutSwapBuffers();
}

/******************************************************************************/
/* end of SELECT                                                              */
/******************************************************************************/



/*********************** ******************************************************/
/* OPTIONS             * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Options screen) ============================================*/
void idleOptions(void)
{
   double frametime = getFrametime();

   if(keys[13] || keys[27])
   {
      changeGameMode(TITLES);
      keys[13] = false;
      keys[27] = false;
   }

   if(boona.msg_counter<4)
      boona.msg_counter+=frametime;
   else
   {
      if(boona.scroll_offset < win.height) boona.scroll_offset+=frametime*25;
      else
      {
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
      }
   }

   glutPostRedisplay();
}

/* Display function (Options screen) =========================================*/
void displayOptions(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);
   drawScene();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   largeOutlineFont("Boona Taxi Options",
                    16, 128-(int)boona.scroll_offset, 0.33333, 0.8, 0.33333);
   smallFont("Please use the Configuration Editor utility to change settings.",
             16, 140-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Alternatively, the following options can be changed in-game...",
             16, 152-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F2",
                    16, 192-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle frames per second counter on/off",
             80, 192-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F3",
                    16, 224-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle heads-up-display stats on/off",
             80, 224-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F4",
                    16, 256-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Change camera view",
             80, 256-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F9",
                    16, 288-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle static lighting on/off",
             80, 288-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F10",
                    16, 320-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle dynamic lighting on/off",
             80, 320-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F11",
                    16, 352-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle environment mapping on/off",
             80, 352-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("F12",
                    16, 384-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Toggle fog on/off",
             80, 384-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   drawOverlay();
   largeFont("Boona Taxi Options Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeOutlineFont("Options", 16, win.height-60, 1.0, 1.0, 1.0);
   sprintf(statusLine, "Press 'ENTER' or 'ESC' to return");

   glutSwapBuffers();
}

/******************************************************************************/
/* end of OPTIONS                                                             */
/******************************************************************************/



/*********************** ******************************************************/
/* HIGHSCORES          * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Highscores screen) =========================================*/
void idleHighscores(void)
{
   double frametime = getFrametime();

   if(keys[13] || keys[27])
   {
      changeGameMode(TITLES);
      keys[13] = false;
      keys[27] = false;
   }

   if(boona.msg_counter<4)
      boona.msg_counter+=frametime;
   else
   {
      if(boona.scroll_offset < win.height) boona.scroll_offset+=frametime*25;
      else
      {
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
      }
   }

   glutPostRedisplay();
}

/* Display function (Highscores screen) ======================================*/
void displayHighscores(void)
{
   char line[16];

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);
   drawScene();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   largeOutlineFont("Name      Car               Score",
                    64, 128-(int)boona.scroll_offset, 0.33333, 0.8, 0.33333);
   for(int i=0; i<10; i++)
   {
      sprintf(line, "%d", i+1);
      largeOutlineFont(line, 16, 160+(i*32)-(int)boona.scroll_offset,
                       1.0, 1-(float)i/10, 0.33333);
      sprintf(line, "%s", boona.highscore[i].name);
      largeOutlineFont(line, 64, 160+(i*32)-(int)boona.scroll_offset,
                       1.0, 1.0, 1.0);
      sprintf(line, "%s", boona.highscore[i].car);
      largeOutlineFont(line, 224, 160+(i*32)-(int)boona.scroll_offset,
                       1.0, 1.0, 1.0);
      sprintf(line, "$%4.2f", boona.highscore[i].score);
      largeOutlineFont(line, 512, 160+(i*32)-(int)boona.scroll_offset,
                       1.0, 1.0, 1.0);
   }

   drawOverlay();
   largeFont("Boona Taxi Highscores Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeOutlineFont("Highscores", 16, win.height-60, 1.0, 1.0, 1.0);
   sprintf(statusLine, "Press 'ENTER' or 'ESC' to return");

   glutSwapBuffers();
}

/******************************************************************************/
/* end of HIGHSCORES                                                          */
/******************************************************************************/



/*********************** ******************************************************/
/* CREDITS             * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Credits screen) ============================================*/
void idleCredits(void)
{
   double frametime = getFrametime();

   if(keys[13] || keys[27])
   {
      changeGameMode(TITLES);
      keys[13] = false;
      keys[27] = false;
   }

   if(boona.msg_counter<4)
      boona.msg_counter+=frametime;
   else
   {
      if(boona.scroll_offset < win.height) boona.scroll_offset+=frametime*25;
      else
      {
         boona.msg_counter   = 0;
         boona.scroll_offset = 0;
      }
   }

   glutPostRedisplay();
}

/* Display function (Credits screen) =========================================*/
void displayCredits(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);
   drawScene();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);

   largeOutlineFont("Written by",
                    16, 128-(int)boona.scroll_offset, 0.33333, 0.8, 0.33333);
   largeOutlineFont("Michael Bonnington",
                    16, 160-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Concept, Coding (C++/OpenGL), Car design, Level design",
             16, 172-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("Enormous thanks to",
                    16, 224-(int)boona.scroll_offset, 0.33333, 0.8, 0.33333);
   largeOutlineFont("Rob Bateman",
                    16, 256-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   smallFont("Fruity game engine, Object loader, Renderer",
             16, 268-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   smallFont("The Author would also like to thank the Boona Racer 2000 team:",
             16, 320-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   largeOutlineFont("Arthur Yarwood",
                    16, 352-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   largeOutlineFont("Eike Anderson",
                    16, 384-(int)boona.scroll_offset, 1.0, 1.0, 1.0);
   largeOutlineFont("JP Li",
                    16, 416-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   largeOutlineFont("Project Tutor",
             16, 480-(int)boona.scroll_offset, 0.33333, 0.8, 0.33333);
   largeOutlineFont("Adam Vanner",
                    16, 512-(int)boona.scroll_offset, 1.0, 1.0, 1.0);

   drawOverlay();
   largeFont("Boona Taxi Credits Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeOutlineFont("Credits", 16, win.height-60, 1.0, 1.0, 1.0);
   sprintf(statusLine, "Press 'ENTER' or 'ESC' to return");


   glutSwapBuffers();
}

/******************************************************************************/
/* end of CREDITS                                                             */
/******************************************************************************/



/*********************** ******************************************************/
/* GAME                * * game mode-specific idle/display callback functions */
/*********************** ******************************************************/

/* Idle function (Game) ======================================================*/
void idleGame(void)
{
   double frametime = getFrametime();

   if(keys['p'] || keys['P'] || keys[27])
   {
      changeGameMode(PAUSED);
      keys['p'] = false;
      keys['P'] = false;
      keys[27] = false;
   }

   if(specials[GLUT_KEY_F2])   //toggle fps counter on/off
   {
      if(boona.show_fps)
      {
         boona.show_fps = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Frames per second counter disabled.");
      }
      else
      {
         boona.show_fps = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Frames per second counter enabled.");
      }
      specials[GLUT_KEY_F2] = false;
   }

   if(specials[GLUT_KEY_F3])   //toggle heads-up-display on/off
   {
      if(boona.show_hud)
      {
         boona.show_hud = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Heads-up-display stats disabled.");
      }
      else
      {
         boona.show_hud = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Heads-up-display stats enabled.");
      }
      specials[GLUT_KEY_F3] = false;
   }

   if(specials[GLUT_KEY_F4])   //change camera view
   {
      boona.cam_counter = 0;
      /*cam.old_pos.x = cam.pos.x;
      cam.old_pos.y = cam.pos.y;
      cam.old_pos.z = cam.pos.z;
      cam.old_aim.x = cam.aim.x;
      cam.old_aim.y = cam.aim.y;
      cam.old_aim.z = cam.aim.z;*/
      cam.old_pos = cam.pos;
      cam.old_aim = cam.aim;
      if(boona.cam_view < 2) boona.cam_view++;
      else boona.cam_view = 0;
      specials[GLUT_KEY_F4] = false;
   }

   if(specials[GLUT_KEY_F5])   //toggle top-down view on/off
   {
      if(boona.top_down_view) boona.top_down_view = false;
      else boona.top_down_view = true;
      specials[GLUT_KEY_F5] = false;
   }

   if(specials[GLUT_KEY_F9])   //toggle static lighting on/off
   {
      if(boona.lighting_s)
      {
         boona.lighting_s = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Static lighting disabled.");
      }
      else
      {
         boona.lighting_s = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Static lighting enabled.");
      }
      specials[GLUT_KEY_F9] = false;
   }

   if(specials[GLUT_KEY_F10])   //toggle dynamic lighting on/off
   {
      if(boona.lighting_d)
      {
         boona.lighting_d = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Dynamic lighting disabled.");
      }
      else
      {
         boona.lighting_d = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Dynamic lighting enabled.");
      }
      specials[GLUT_KEY_F10] = false;
   }

   if(specials[GLUT_KEY_F11])   //toggle environment mapping on/off
   {
      if(boona.use_env_map)
      {
         boona.use_env_map = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Environment mapping disabled.");
      }
      else
      {
         boona.use_env_map = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Environment mapping enabled.");
      }
      specials[GLUT_KEY_F11] = false;
   }

   if(specials[GLUT_KEY_F12])   //toggle fog on/off
   {
      if(boona.use_fog)
      {
         boona.use_fog = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Fog disabled.");
      }
      else
      {
         boona.use_fog = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Fog enabled.");
      }
      specials[GLUT_KEY_F12] = false;
   }


   /* Test player position against level array (blag alert)... */
   if(level_map[(int)floor(player.pos_x)][(int)floor(player.pos_z)])
   {
      player.setPos(player.pos_x-player.dx,
                    0,
                    player.pos_z-player.dz,
                    player.getSpeed()/4,
                    player.getAngle(),
                    player.getHeading());
cout << "BANG!" << endl;
   }
   else
   {
      /* Update player position... */
      player.update(frametime);

      /* Update camera position... */
      if(boona.cam_counter < 1-frametime*2)
      {
         boona.cam_counter+=frametime*2;
         updateCam();
      }
      else
      {
         boona.cam_counter = 1;
         positionCam();
      }
   }


   /* Update game time... */
   if(boona.game_time > 0) boona.game_time -= frametime;
   else
   {
      boona.game_time = 0;
      changeGameMode(TITLES);
   }

   if(boona.msg_counter<2)
      boona.msg_counter+=frametime;
   else
      if(boona.show_fps)
         sprintf(statusLine, "Frames/second: %6.2f", (float)1/frametime);
      else
         sprintf(statusLine, "");

   glutPostRedisplay();
}

/* Display function (Game) ===================================================*/
void displayGame(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);
   drawScene();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);
   smallFont(statusLine, 16, win.height-16, 1.0, 1.0, 1.0);
   if(boona.show_hud) drawHUD();

   glutSwapBuffers();
}

/******************************************************************************/
/* end of GAME                                                                */
/******************************************************************************/



/******************************************************************************/
/* PAUSED game mode-specific GLUT callback function:                          */
/*        idle, display                                                       */
/******************************************************************************/

/* Idle function (Paused screen) =============================================*/
void idlePaused(void)
{
   double frametime = getFrametime();

   if(keys['p'] || keys['P'] || keys[13])
   {
      changeGameMode(GAME);
      keys['p'] = false;
      keys['P'] = false;
      keys[13] = false;
   }

   if(keys[27])
   {
      changeGameMode(TITLES);
      keys[27] = false;
   }

   if(specials[GLUT_KEY_F9])   //toggle static lighting on/off
   {
      if(boona.lighting_s)
      {
         boona.lighting_s = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Static lighting disabled.");
      }
      else
      {
         boona.lighting_s = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Static lighting enabled.");
      }
      specials[GLUT_KEY_F9] = false;
   }

   if(specials[GLUT_KEY_F10])   //toggle dynamic lighting on/off
   {
      if(boona.lighting_d)
      {
         boona.lighting_d = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Dynamic lighting disabled.");
      }
      else
      {
         boona.lighting_d = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Dynamic lighting enabled.");
      }
      specials[GLUT_KEY_F10] = false;
   }

   if(specials[GLUT_KEY_F11])   //toggle environment mapping on/off
   {
      if(boona.use_env_map)
      {
         boona.use_env_map = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Environment mapping disabled.");
      }
      else
      {
         boona.use_env_map = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Environment mapping enabled.");
      }
      specials[GLUT_KEY_F11] = false;
   }

   if(specials[GLUT_KEY_F12])   //toggle fog on/off
   {
      if(boona.use_fog)
      {
         boona.use_fog = false;
         boona.msg_counter = 0;
         sprintf(statusLine, "Fog disabled.");
      }
      else
      {
         boona.use_fog = true;
         boona.msg_counter = 0;
         sprintf(statusLine, "Fog enabled.");
      }
      specials[GLUT_KEY_F12] = false;
   }

   sprintf(statusLine, "Press 'P' to resume or 'ESC' to quit");

   glutPostRedisplay();
}

/* Display function (Paused screen) ==========================================*/
void displayPaused(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   /* Draw all the 3D stuff... */
   glCallList(persp);

   glPushMatrix();
    glColor3f(1.0, 1.0, 1.0);
    glTranslatef(boona.trans_x, boona.trans_y, boona.dolly);
    glRotatef(boona.spin_x, 1.0, 0.0, 0.0);

    glPushMatrix();
     glRotatef(boona.spin_y, 0.0, 1.0, 0.0);
     drawScene();
    glPopMatrix();

   glPopMatrix();

   /* Draw all the 2D stuff on top of the 3D... */
   glCallList(ortho);
   drawOverlay();
   largeFont("Boona Taxi Paused Screen", 16, 48, 0.93333, 0.66667, 0.0);
   largeOutlineFont("Game Paused", 16, win.height-60, 1.0, 1.0, 1.0);

   glutSwapBuffers();
}

/******************************************************************************/
/* end of PAUSED                                                              */
/******************************************************************************/



/* Reshape function ==========================================================*/
void reshape(int w, int h)
{
   if(h == 0) h = 1;   //Required to prevent crash when window is minimised

   win.width = w;
   win.height = h;

   glDeleteLists(persp, 1);
   glDeleteLists(ortho, 1);
   persp = glGenLists(1);
   ortho = glGenLists(1);

   glNewList(persp, GL_COMPILE);
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(75.0, (GLfloat) w/(GLfloat) h, 0.1, FARPLANE);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
   glEndList();

   glNewList(ortho, GL_COMPILE);
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, w, h, 0, -1, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);
   glEndList();

   glCallList(persp);
}

/* Keyboard function - key is pressed ========================================*/
void keyDown(unsigned char key, int x, int y)
{
   keys[key] = true;

   if(boona.game_mode == GAME)
   {
      switch(key)
      {
         case 'z':
            player.setBrake(true);
            break;

         case 'x':
            player.setAccelerator(true);
            break;

         case ' ':
            player.setHandbrake(true);
            break;

         default:
            break;
      }
   }
}

/* Keyboard function - key is released =======================================*/
void keyUp(unsigned char key, int x, int y)
{
   keys[key] = false;

   if(boona.game_mode == GAME)
   {
      switch(key)
      {
         case 'z':
            player.setBrake(false);
            break;

         case 'x':
            player.setAccelerator(false);
            break;

         case ' ':
            player.setHandbrake(false);
            break;

         default:
            break;
      }
   }
}

/* Keyboard function - special key is pressed ================================*/
void specialDown(int key, int x, int y)
{
   specials[key] = true;

   if(boona.game_mode == GAME)
   {
      switch(key)
      {
         case GLUT_KEY_LEFT:
            player.setLeft(true);
            player.setRight(false);
            break;

         case GLUT_KEY_RIGHT:
            player.setRight(true);
            player.setLeft(false);
            break;

         case GLUT_KEY_UP:
            player.setReverse(false);
            break;

         case GLUT_KEY_DOWN:
            player.setReverse(true);
            break;

         default:
            break;
      }
   }
}

/* Keyboard function - special key is released ===============================*/
void specialUp(int key, int x, int y)
{
   specials[key] = false;

   if(boona.game_mode == GAME)
   {
      switch(key)
      {
         case GLUT_KEY_LEFT:
            player.setLeft(false);
            break;

         case GLUT_KEY_RIGHT:
            player.setRight(false);
            break;

         default:
            break;
      }
   }
}

/* Mouse input function ======================================================*/
void mouse(int button, int state, int x, int y)
{
   if(state == GLUT_DOWN)
   {
      boona.mouse_press_x = x; boona.mouse_press_y = y;

      switch(button)
      {
         case GLUT_LEFT_BUTTON:
            boona.lmb = true;
            break;

         case GLUT_MIDDLE_BUTTON:
            boona.mmb = true;
            break;

         default:
            break;
      }
   }

   if(state == GLUT_UP)
   {
      glutSetCursor(GLUT_CURSOR_NONE);
      switch(button)
      {
         case GLUT_LEFT_BUTTON:
            boona.lmb = false;
            break;

         case GLUT_MIDDLE_BUTTON:
            boona.mmb = false;
            break;

         default:
            break;
      }
   }

   glutPostRedisplay();
}

/* Motion function ===========================================================*/
void motion(int x, int y)
{
   int step_x = (x - boona.mouse_press_x), step_y = (y - boona.mouse_press_y);

   boona.mouse_press_x = x; boona.mouse_press_y = y;

   if(boona.lmb && !boona.mmb)        /* Tumble */
   {
#ifdef WIN32
      glutSetCursor(GLUT_CURSOR_CROSSHAIR);
#else
      glutSetCursor(GLUT_CURSOR_CYCLE);
#endif
      boona.spin_x += 0.5*step_y;
      boona.spin_y += 0.5*step_x;
   }
   else if(boona.mmb && !boona.lmb)   /* Translate */
   {
#ifdef WIN32
      glutSetCursor(GLUT_CURSOR_SPRAY);
#else
      glutSetCursor(GLUT_CURSOR_CROSSHAIR);
#endif
      boona.trans_x += 0.03*step_x;
      boona.trans_y -= 0.03*step_y;
   }
   else if(boona.lmb && boona.mmb)    /* Dolly */
   {
      glutSetCursor(GLUT_CURSOR_LEFT_RIGHT);
      boona.dolly += 0.03*step_x;
   }
   glutPostRedisplay();
}
