/******************************************************************************/
/*                                                                            */
/*        Project: Boona Taxi                                                 */
/*          Title: Configuration editor (using GLUI)                          */
/*        Version: 0.3                                                        */
/*                                                                            */
/*        Started: 28/02/2001                                                 */
/*   Last Updated: 22/03/2001                                                 */
/*                                                                            */
/*         Author: Michael Bonnington                                         */
/*                                                                            */
/******************************************************************************/

#include <string.h>
#include <fstream.h>

#include "./opengl_includes.h"
#include "./font.h"

#ifdef WIN32
   #include <GL/glui.h>
#else
   #include "./glui.h"
#endif


#define CREATE_WINDOW 0
#define SET_WINDOW 1
#define SET_SCREENMODE 2
#define SET_FOG 13
#define SET_RESOLUTION 3
#define SET_ASPECT_RATIO 4
#define ADJUST_X_RESOLUTION 5
#define ADJUST_Y_RESOLUTION 6
#define PRESET_FASTEST 7
#define PRESET_BEST 8
#define INIT 9
#define RESET 10
#define SAVE 11
#define EXIT 12



/* Function prototypes =======================================================*/
void control(int);
void createButtonsSubwindow(void);
void createWindowOptionsSubwindow(void);
void createRenderOptionsSubwindow(void);
void createGameOptionsSubwindow(void);



/* Globals ===================================================================*/
int main_window;


/* These are the live variables passed into GLUI... */
int window_string = 0, current_window = 0,
    fullscreen,
    resolution_string = 6,
    resolution_x, resolution_y,
    window_pos_x, window_pos_y,
    bit_depth,
    refresh_rate,
    preserve_aspect_ratio = 1,
    lighting_s,
    lighting_d,
    env_map,
    fog,
    texture_filter,
    time_limit,
    show_hud,
    show_fps,
    cam_view;

GLfloat fog_colour[4] = {0.6, 0.8, 1.0, 1.0};
float aspect_ratio;

char *window_string_list[] = {"Window modes",
                              "Rendering",
                              "Gameplay"};

char *resolution_string_list[] = {"400x300",
                                  "640x480",
                                  "800x600",
                                  "1024x768",
                                  "1280x1024",
                                  "1600x1200",
                                  "Custom..."};


/* Pointers to GLUI subwindows... */
GLUI *buttons, *window_options, *render_options, *game_options;


/* Pointers to GLUI controls... */
GLUI_Panel *resolution_panel, *window_panel, *fullscreen_panel,
           *fog_colour_panel;

GLUI_Listbox *resolution_listbox;

GLUI_Spinner *resolution_x_spinner, *resolution_y_spinner,
             *window_pos_x_spinner, *window_pos_y_spinner,
             *refresh_rate_spinner,
             *fog_r_spinner, *fog_g_spinner, *fog_b_spinner, *fog_a_spinner,
             *time_limit_spinner;

GLUI_Checkbox *aspect_ratio_checkbox,
              *lighting_s_checkbox,
              *lighting_d_checkbox,
              *env_map_checkbox,
              *fog_checkbox,
              *show_hud_checkbox,
              *show_fps_checkbox;

GLUI_RadioGroup *fullscreen_radio, *bit_depth_radio,
                *texture_radio,
                *cam_view_radio;

/* Strings to hold the sample text file... */
char line1[20],
     line2[20],
     line3[20],
     line4[20],
     line5[20],
     line6[20],
     line7[20];



/* Functions =================================================================*/

/* Load config data ==========================================================*/
bool loadData(void)
{
   ifstream window_opt,
            render_opt,
            game_opt;
   char in_str[80];
   bool success = true;


   /* Read in "WindowOptions.cfg" file... */
   window_opt.open("./config/WindowOptions.cfg", ios::in);

   if(window_opt.fail()) success = false;
   else   //load values from config file
   {
      window_opt >> in_str >> resolution_x;
      window_opt >> in_str >> resolution_y;
      window_opt >> in_str >> window_pos_x;
      window_opt >> in_str >> window_pos_y;

      window_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) fullscreen = 1;
      else fullscreen = 0;

      window_opt >> in_str >> in_str;
      if(strcmp("32", in_str) == 0) bit_depth = 1;
      else bit_depth = 0;

      window_opt >> in_str >> refresh_rate;

      window_opt.close();
   }


   /* Read in "RenderingOptions.cfg" file... */
   render_opt.open("./config/RenderingOptions.cfg", ios::in);

   if(render_opt.fail()) success = false;
   else   //load values from config file
   {
      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) lighting_s = 1;
      else lighting_s = 0;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) lighting_d = 1;
      else lighting_d = 0;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) env_map = 1;
      else env_map = 0;

      render_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) fog = 1;
      else fog = 0;

      render_opt >> in_str
                 >> fog_colour[0]
                 >> fog_colour[1]
                 >> fog_colour[2]
                 >> fog_colour[3];

      render_opt >> in_str >> texture_filter;

      render_opt.close();
   }


   /* Read in "GameOptions.cfg" file... */
   game_opt.open("./config/GameOptions.cfg", ios::in);

   if(game_opt.fail()) success = false;
   else   //load values from config file
   {
      game_opt >> in_str >> time_limit;
      time_limit = (int)time_limit/60;   //convert seconds to minutes

      game_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) show_hud = 1;
      else show_hud = 0;

      game_opt >> in_str >> in_str;
      if(strcmp("TRUE", in_str) == 0) show_fps = 1;
      else show_fps = 0;

      game_opt >> in_str >> cam_view;

      game_opt.close();
   }


   return success;
}

/* Save config data ==========================================================*/
bool saveData(void)
{
   ofstream window_opt,
            render_opt,
            game_opt;
   char out_str[80];
   bool success = true;


   /* Save out "WindowOptions.cfg" file... */
   window_opt.open("./config/WindowOptions.cfg", ios::out);

   if(window_opt.fail()) success = false;
   else   //save values to config file
   {
      window_opt << "SCREEN_WIDTH " << resolution_x << endl;
      window_opt << "SCREEN_HEIGHT " << resolution_y << endl;
      window_opt << "SCREEN_POS_X " << window_pos_x << endl;
      window_opt << "SCREEN_POS_Y " << window_pos_y << endl;

      if(fullscreen == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      window_opt << "FULLSCREEN " << out_str << endl;

      if(bit_depth == 1) sprintf(out_str, "32");
      else sprintf(out_str, "16");
      window_opt << "BIT_DEPTH " << out_str << endl;

      window_opt << "REFRESH_RATE " << refresh_rate << endl;

      window_opt.close();
   }


   /* Save out "RenderingOptions.cfg" file... */
   render_opt.open("./config/RenderingOptions.cfg", ios::out);

   if(render_opt.fail()) success = false;
   else   //load values from config file
   {
      if(lighting_s == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      render_opt << "STATIC_LIGHTING " << out_str << endl;

      if(lighting_d == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      render_opt << "DYNAMIC_LIGHTING " << out_str << endl;

      if(env_map == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      render_opt << "ENVIRONMENT_MAPPING " << out_str << endl;

      if(fog == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      render_opt << "FOG " << out_str << endl;

      render_opt << "FOG_COLOUR "
                 << fog_colour[0] << " "
                 << fog_colour[1] << " "
                 << fog_colour[2] << " "
                 << fog_colour[3]
                 << endl;

      render_opt << "TEXTURE_FILTER " << texture_filter << endl;

      render_opt.close();
   }


   /* Save out "GameOptions.cfg" file... */
   game_opt.open("./config/GameOptions.cfg", ios::out);

   if(game_opt.fail()) success = false;
   else   //save values to config file
   {
      game_opt << "TIME_LIMIT " << (int)time_limit*60 << endl;

      if(show_hud == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      game_opt << "SHOW_HUD " << out_str << endl;

      if(show_fps == 1) sprintf(out_str, "TRUE");
      else sprintf(out_str, "FALSE");
      game_opt << "SHOW_FPS " << out_str << endl;

      game_opt << "CAMERA_VIEW " << cam_view << endl << endl;

      game_opt << "STEER_LEFT" << endl;
      game_opt << "STEER_RIGHT" << endl;
      game_opt << "SHIFT_UP" << endl;
      game_opt << "SHIFT_DOWN" << endl;
      game_opt << "BRAKE" << endl;
      game_opt << "ACCELERATE" << endl;
      game_opt << "HANDBRAKE" << endl;

      game_opt.close();
   }


   return success;
}

/* Set a new resolution ======================================================*/
void setResolution(int x, int y)
{
   resolution_x_spinner->set_int_val(x);
   resolution_y_spinner->set_int_val(y);
   aspect_ratio = (float)x/y;
}

/* Control callback function to act on user input ============================*/
void control(int instruction)
{
   switch(instruction)
   {
      case CREATE_WINDOW:   //Choose which options window to display...
         switch(current_window)
         {
            case 0:
               window_options->close();
               control(SET_WINDOW);
               break;

            case 1:
               render_options->close();
               control(SET_WINDOW);
               break;

            case 2:
               game_options->close();
               control(SET_WINDOW);
               break;

            default:
               break;
         }
         break;

      case SET_WINDOW:   //Choose which options window to display...
         switch(window_string)
         {
            case 0:
               createWindowOptionsSubwindow();
               control(INIT);
#ifdef WIN32
#else
               //Bug work-around re-creates the buttons subwindow in UNIX
               buttons->close();
               createButtonsSubwindow();
               buttons->sync_live();
#endif
               current_window = window_string;
               break;

            case 1:
               createRenderOptionsSubwindow();
#ifdef WIN32
#else
               //Bug work-around re-creates the buttons subwindow in UNIX
               buttons->close();
               createButtonsSubwindow();
               buttons->sync_live();
#endif
               current_window = window_string;
               break;

            case 2:
               createGameOptionsSubwindow();
#ifdef WIN32
#else
               //Bug work-around re-creates the buttons subwindow in UNIX
               buttons->close();
               createButtonsSubwindow();
               buttons->sync_live();
#endif
               current_window = window_string;
               break;

            default:
               break;
         }
         break;

      case SET_SCREENMODE:   //Ensure correct panels are enabled...
         if(fullscreen)
         {
            window_panel->disable();
            fullscreen_panel->enable();
         }
         else
         {
            window_panel->enable();
            fullscreen_panel->disable();
         }
         break;

      case SET_FOG:   //Ensure correct panels are enabled...
         if(fog)
            fog_colour_panel->enable();
         else
            fog_colour_panel->disable();
         glutPostRedisplay();
         break;

      case SET_RESOLUTION:   //Set resolution using listbox...
         if(resolution_string == 6)
            resolution_panel->enable();
         else
         {
            resolution_panel->disable();

            switch(resolution_string)
            {
               case 0:
                  setResolution(400, 300);
                  break;

               case 1:
                  setResolution(640, 480);
                  break;

               case 2:
                  setResolution(800, 600);
                  break;

               case 3:
                  setResolution(1024, 768);
                  break;

               case 4:
                  setResolution(1280, 1024);
                  break;

               case 5:
                  setResolution(1600, 1200);
                  break;

               default:
                  setResolution(640, 480);
                  break;
            }
         }
         break;

      case SET_ASPECT_RATIO:   //Calculate aspect ratio...
         aspect_ratio = (float)resolution_x/resolution_y;
         break;

      case ADJUST_X_RESOLUTION:   //Adjust x resolution using spinner...
         if(preserve_aspect_ratio)
            resolution_y_spinner->set_int_val(resolution_x/aspect_ratio);
         else
            aspect_ratio = (float)resolution_x/resolution_y;
         break;

      case ADJUST_Y_RESOLUTION:   //Adjust y resolution using spinner...
         if(preserve_aspect_ratio)
            resolution_x_spinner->set_int_val(resolution_y*aspect_ratio);
         else
            aspect_ratio = (float)resolution_x/resolution_y;
         break;

      case PRESET_FASTEST:   //Set preset values for fastest render options...
         lighting_s = 0;
         lighting_d = 0;
         env_map = 0;
         fog = 0;
         texture_filter = 0;
         render_options->sync_live();
         break;

      case PRESET_BEST:   //Set preset values for best render options...
         lighting_s = 1;
         lighting_d = 1;
         env_map = 1;
         fog = 1;
         texture_filter = 2;
         render_options->sync_live();
         break;

      case INIT:   //Reset interface values...
         control(SET_SCREENMODE);
         control(SET_RESOLUTION);
         window_options->sync_live();
         break;

      case RESET:   //Reset all values to their defaults...
         fullscreen = 0;
         resolution_string = 6;
         resolution_x = 640;
         resolution_y = 480;
         window_pos_x = 100;
         window_pos_y = 100;
         bit_depth = 0;
         refresh_rate = 60;
         lighting_s = 1;
         lighting_d = 1;
         env_map = 1;
         fog = 1;
         fog_colour[0] = 0.6;
         fog_colour[1] = 0.8;
         fog_colour[2] = 1.0;
         fog_colour[3] = 1.0;
         texture_filter = 1;
         time_limit = 5;
         show_hud = 1;
         show_fps = 0;
         cam_view = 1;
         GLUI_Master.sync_live_all();
         break;

      case SAVE:   //Accept (save data to config files)...
         if(!saveData())
            printf("\nERROR: Unable to save configuration files\n");
         else
            printf("\nConfiguration files saved sucessfully\n");
         control(EXIT);
         break;

      case EXIT:   //Cancel (exit)...
         GLUI_Master.close_all();
         exit(0);
         break;

      default:
         break;
   }
}

/* Function to create buttons subwindow ======================================*/
void createButtonsSubwindow(void)
{
   buttons =
      GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_BOTTOM);

   /* Link window to GLUI... */
   buttons->set_main_gfx_window(main_window);


   GLUI_Panel *buttons_panel = buttons->add_panel("Buttons", GLUI_PANEL_NONE);

   /* Listbox to select which options window to view... */
   GLUI_Listbox *window_listbox =
      buttons->add_listbox_to_panel(buttons_panel, "Display options for:",
                                    &window_string, CREATE_WINDOW, control);
   for(int i=0; i<3; i++)
      window_listbox->add_item(i, window_string_list[i]);

   buttons->add_column_to_panel(buttons_panel, false);
   buttons->add_column_to_panel(buttons_panel, true);
   buttons->add_column_to_panel(buttons_panel, false);

   /* Defaults, Accept and Cancel buttons... */
   buttons->add_button_to_panel(buttons_panel, "Defaults", RESET, control);
   buttons->add_column_to_panel(buttons_panel, false);
   buttons->add_button_to_panel(buttons_panel, "Save", SAVE, control);
   buttons->add_column_to_panel(buttons_panel, false);
   buttons->add_button_to_panel(buttons_panel, "Cancel", EXIT, control);

   buttons_panel->set_alignment(GLUI_ALIGN_RIGHT);   //doesn't seem to work
}

/* Function to create window_options subwindow ===============================*/
void createWindowOptionsSubwindow(void)
{
   window_options =
      GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_LEFT);

   /* Link window to GLUI... */
   window_options->set_main_gfx_window(main_window);


   /* Radio buttons to change between fullscreen and windowed mode... */
   fullscreen_radio =
      window_options->add_radiogroup(&fullscreen,
                                     SET_SCREENMODE, control);
   window_options->add_radiobutton_to_group(fullscreen_radio, "Window");
   window_options->add_radiobutton_to_group(fullscreen_radio, "Fullscreen");


   window_options->add_separator();


   /* Listbox to change resolutions... */
   resolution_listbox =
      window_options->add_listbox("Resolution:", &resolution_string,
                                  SET_RESOLUTION, control);
   for(int i=0; i<7; i++)
      resolution_listbox->add_item(i, resolution_string_list[i]);

   /* Resolution options panel... */
   resolution_panel =
      window_options->add_panel("Resolution options", GLUI_PANEL_NONE);

   /* Spinners to adjust resolution... */
   resolution_x_spinner =
      window_options->add_spinner_to_panel(resolution_panel, "x:",
                                           GLUI_SPINNER_INT, &resolution_x,
                                           ADJUST_X_RESOLUTION, control);
   resolution_x_spinner->set_int_limits(320, 3200);
   resolution_y_spinner =
      window_options->add_spinner_to_panel(resolution_panel, "y:",
                                           GLUI_SPINNER_INT, &resolution_y,
                                           ADJUST_Y_RESOLUTION, control);
   resolution_y_spinner->set_int_limits(240, 2400);

   aspect_ratio_checkbox =
      window_options->add_checkbox_to_panel(resolution_panel,
                                            "Preserve aspect ratio",
                                            &preserve_aspect_ratio);


   window_options->add_separator();


   /* Window-specific options panel... */
   window_panel =
      window_options->add_panel("Window-specific options", GLUI_PANEL_NONE);

   /* Spinners to adjust window position... */
   window_options->add_statictext_to_panel(window_panel, "Window position:");
   window_pos_x_spinner =
      window_options->add_spinner_to_panel(window_panel, "x:",
                                           GLUI_SPINNER_INT, &window_pos_x);
   window_pos_x_spinner->set_int_limits(1, resolution_x);
   window_pos_y_spinner =
      window_options->add_spinner_to_panel(window_panel, "y:",
                                           GLUI_SPINNER_INT, &window_pos_y);
   window_pos_y_spinner->set_int_limits(1, resolution_y);


   window_options->add_separator();


   /* Fullscreen-specific options panel... */
   fullscreen_panel =
      window_options->add_panel("Fullscreen-specific options", GLUI_PANEL_NONE);

   /* Radio buttons to change bit depth... */
   window_options->add_statictext_to_panel(fullscreen_panel, "Bit depth:");
   bit_depth_radio =
      window_options->add_radiogroup_to_panel(fullscreen_panel, &bit_depth);
   window_options->add_radiobutton_to_group(bit_depth_radio, "16-bit");
   window_options->add_radiobutton_to_group(bit_depth_radio, "32-bit");

   /* Spinner to adjust refresh rate... */
   refresh_rate_spinner =
      window_options->add_spinner_to_panel(fullscreen_panel,
                                           "Refresh rate (Hz):",
                                           GLUI_SPINNER_INT,
                                           &refresh_rate);
   refresh_rate_spinner->set_int_limits(60, 240);
}

/* Function to create render_options subwindow ===============================*/
void createRenderOptionsSubwindow(void)
{
   render_options =
      GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_LEFT);

   /* Link window to GLUI... */
   render_options->set_main_gfx_window(main_window);


   /* Image quality options panel... */
   GLUI_Panel *image_panel =
      render_options->add_panel("Image quality");

   lighting_s_checkbox =
      render_options->add_checkbox_to_panel(image_panel,
                                            "Static lighting (pre-calculated)",
                                            &lighting_s);
   lighting_d_checkbox =
      render_options->add_checkbox_to_panel(image_panel,
                                            "Dynamic lighting (OpenGL)",
                                            &lighting_d);
   env_map_checkbox =
      render_options->add_checkbox_to_panel(image_panel,
                                            "Environment mapping",
                                            &env_map);
   fog_checkbox =
      render_options->add_checkbox_to_panel(image_panel,
                                            "Fog",
                                            &fog,
                                            SET_FOG, control);
   image_panel->set_alignment(GLUI_ALIGN_LEFT);

   /* Fog colour options panel... */
   fog_colour_panel =
      render_options->add_panel("Fog colour");
   fog_r_spinner =
      render_options->add_spinner_to_panel(fog_colour_panel, "Red:",
                                           GLUI_SPINNER_FLOAT, &fog_colour[0]);
   fog_r_spinner->set_float_limits(0.0, 1.0);
   fog_g_spinner =
      render_options->add_spinner_to_panel(fog_colour_panel, "Green:",
                                           GLUI_SPINNER_FLOAT, &fog_colour[1]);
   fog_g_spinner->set_float_limits(0.0, 1.0);
   fog_b_spinner =
      render_options->add_spinner_to_panel(fog_colour_panel, "Blue:",
                                           GLUI_SPINNER_FLOAT, &fog_colour[2]);
   fog_b_spinner->set_float_limits(0.0, 1.0);
   fog_a_spinner =
      render_options->add_spinner_to_panel(fog_colour_panel, "Alpha:",
                                           GLUI_SPINNER_FLOAT, &fog_colour[3]);
   fog_a_spinner->set_float_limits(0.0, 1.0);
   fog_colour_panel->set_alignment(GLUI_ALIGN_LEFT);

   /* Texture quality options panel... */
   GLUI_Panel *texture_panel =
      render_options->add_panel("Texture filter quality");
   texture_radio =
      render_options->add_radiogroup_to_panel(texture_panel, &texture_filter);
   render_options->add_radiobutton_to_group(texture_radio,
                                            "Nearest neighbour");
   render_options->add_radiobutton_to_group(texture_radio,
                                            "Bilinear");
   render_options->add_radiobutton_to_group(texture_radio,
                                            "Bilinear with mipmapping");
   texture_panel->set_alignment(GLUI_ALIGN_LEFT);

   /* Presets buttons panel... */
   GLUI_Panel *presets_panel =
      render_options->add_panel("Presets");
   GLUI_Button *preset_fastest_button =
      render_options->add_button_to_panel(presets_panel,
                                          "Fastest (most compatible)",
                                          PRESET_FASTEST, control);
   preset_fastest_button->set_w(160);
   GLUI_Button *preset_best_button =
      render_options->add_button_to_panel(presets_panel,
                                          "Best image quality",
                                          PRESET_BEST, control);
   preset_best_button->set_w(160);
   presets_panel->set_alignment(GLUI_ALIGN_LEFT);
}

/* Function to create game_options subwindow =================================*/
void createGameOptionsSubwindow(void)
{
   game_options =
      GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_LEFT);

   /* Link window to GLUI... */
   game_options->set_main_gfx_window(main_window);


   /* Spinner to change time limit... */
   time_limit_spinner =
      game_options->add_spinner("Time limit (minutes):", GLUI_SPINNER_INT,
                                &time_limit);
   time_limit_spinner->set_int_limits(1, 15);


   game_options->add_separator();


   /* Checkbox for HUD... */
   show_hud_checkbox =
      game_options->add_checkbox("Show heads-up display stats",
                                 &show_hud);

   /* Checkbox for FPS counter... */
   show_fps_checkbox =
      game_options->add_checkbox("Show frames per second counter",
                                 &show_fps);


   game_options->add_separator();


   /* Default camera view radio group... */
   GLUI_Panel *cam_view_panel =
      game_options->add_panel("Default camera view");
   cam_view_radio =
      game_options->add_radiogroup_to_panel(cam_view_panel, &cam_view);
   game_options->add_radiobutton_to_group(cam_view_radio, "In-car");
   game_options->add_radiobutton_to_group(cam_view_radio, "Chase");
   game_options->add_radiobutton_to_group(cam_view_radio, "Helicopter");
}



/* Idle function =============================================================*/
void idle(void)
{
   /* According to the GLUT specification, the current window is undefined
    * during an idle callback. So we need to explicitly change it if necessary.
    */

   if(glutGetWindow() != main_window)
      glutSetWindow(main_window);


   sprintf(line1, "SCREEN_WIDTH %d",  resolution_x);
   sprintf(line2, "SCREEN_HEIGHT %d", resolution_y);
   sprintf(line3, "SCREEN_POS_X %d",  window_pos_x);
   sprintf(line4, "SCREEN_POS_Y %d",  window_pos_y);
   sprintf(line5, "FULLSCREEN %d",    fullscreen);
   sprintf(line6, "TEXTURE_FILTER %d", texture_filter);
   sprintf(line7, "CAMERA_VIEW %d",    cam_view);

   glutPostRedisplay();
}

/* Display function ==========================================================*/
void display(void)
{
   glClearColor(fog_colour[0],
                fog_colour[1],
                fog_colour[2],
                fog_colour[3]);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   smallFont(line1, 256, 32,  0.0, 0.0, 0.0);
   smallFont(line2, 256, 48,  0.0, 0.0, 0.0);
   smallFont(line3, 256, 64,  0.0, 0.0, 0.0);
   smallFont(line4, 256, 80,  0.0, 0.0, 0.0);
   smallFont(line5, 256, 96,  0.0, 0.0, 0.0);
   smallFont(line6, 256, 112, 0.0, 0.0, 0.0);
   smallFont(line7, 256, 128, 0.0, 0.0, 0.0);

   glutSwapBuffers();
}

/* Reshape function ==========================================================*/
void reshape(int w, int h)
{
   if(h == 0) h = 1;   //Required to prevent crash when window is minimised

   //win.width = w;
   //win.height = h;

   int tx, ty, tw, th;
   GLUI_Master.get_viewport_area(&tx, &ty, &tw, &th);
   glViewport(tx, ty, (GLsizei) tw, (GLsizei) th);
   //glViewport(0, 0, (GLsizei) w, (GLsizei) h);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0, tw, th, 0, -1, 1);
   //glOrtho(0, w, h, 0, -1, 1);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}



/* Main function =============================================================*/
void main(int argc, char* argv[])
{
   /* Initialize GLUT and create window... */
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
   glutInitWindowPosition(100, 100);
   glutInitWindowSize(640, 480);

   main_window = glutCreateWindow("Boona Taxi Configuration Editor");
   glutDisplayFunc(display);
   GLUI_Master.set_glutReshapeFunc(reshape);

   makeRasterFont();


   /* Load data from config files... */
   if(!loadData())
   {
      printf("\nUnable to load configuration files - using default values\n");
      control(RESET);   //set default values
   }
   else
      printf("\nConfiguration files loaded sucessfully\n");


   createButtonsSubwindow();
   control(SET_WINDOW);
   //control(INIT);


   /* We register the idle callback with GLUI, *not* with GLUT... */
   GLUI_Master.set_glutIdleFunc(idle);


   glutMainLoop();
}
