/*#####################################################################################
 *
 *   Title        : fruity_matrix.h
 *   Author       : Robert Bateman
 *   Date         : 23-Nov-2000
 *   Last Updated : 05-Dec-2000
 *
 *   Description  :
 *
 *                  A class to provide matrix based maths. Works nicely with cVector3f.
 *   This will help perform all concatonated matrix multiplications needed for 
 *   hierarchical animation + any other uses I can think of.....
 *
#####################################################################################*/

/*
 * Revision History
 *
 * 06/01/2001 : Merged all class definitions into fruity_math.h
 *
 * 05/12/2000 : Used lots of returned 'matrix& func(const matrix)' constructions to
 *              allow you to append transforms onto the back of others. Included a 
 *              rotate around axis function. axis provided by vector, rot by float.
 *              Hopefully no future ammendments should need to be made..............
 *
 * 02/12/2000 : Sorted Interaction with cVector3f class. Dont know if its worth doing
 *              Any multiplication interactions with planes, we'll wait and see.
 *
 * 30/11/2000 : Changed around transformation stuff. Realised you had to perform lots
 *              of matrix multiplications to get it to work.
 *
*/

/*##############################################################################
###                        pre-processor gubbins
##############################################################################*/

#ifndef FRUITYMATH_H
#define FRUITYMATH_H

#define PI_OVER_180 0.017453293
#define PI 3.14159265

#include <math.h>

/*##############################################################################
###                           VERTEX3F DATA CLASS
##############################################################################*/
/*
 *   ****************************** Usage : ********************************
 *
 *   float val = 1;
 *   cVector3f a,b;        : vector class declaration
 *
 *
 *   // special vector functions
 *
 *   a * b;               : 'a' dot 'b'
 *   a ^ b;               : 'a' cross 'b'
 *   val = a.length();    : returns magnitude of vector 'a'
 *   a.normalise();       : makes magnitude of vector 'a' = 1
 *   b = a.normalised();  : makes 'b' = vector 'a' with a magnitude of 1
 *
 *
 *   // vector : vector maths
 *
 *   a + b;               : vector addition
 *   a += b;              : increments x,y,z values of 'a' by x,y,z of 'b'
 *   a - b;               : vector subtraction
 *   a -= b;              : decrements x,y,z values of 'a' by x,y,z of 'b'
 *   a = b;               : makes x,y,z of 'a' = x,y,z of 'b'
 *
 *
 *   // vector : constant maths
 *
 *   a - val;             : subtracts 'val' from x,y & z of 'a'
 *   a + val;             : adds 'val' to x,y & z of 'a'
 *   a += val;            : increments x,y,z of 'a' by 'val'
 *   a -= val;            : decrements x,y,z of 'a' by 'val'
 *   a * val;             : scales x,y,z of 'a' by 'val'
 *   a / val;             : scales x,y,z of 'a' by '1/val'
*/

class cVector3f
  {

    public :

      /*------------------------------ DATA --------------------------------*/

      float x;
      float y;
      float z;
      float w;

      inline void set(float X,float Y,float Z) 
        {
           x=X; y=Y; z=Z;
        };

      /*------------------------ DOT & CROSS PRODUCTS ----------------------*/

      // dot product
      inline float operator*( const cVector3f &v )
        { 
          return ( (v.x*x) + (v.y*y) + (v.z*z) );
        };

      // cross product
      inline cVector3f operator^( const cVector3f &v )
        {
          return cVector3f( (y*v.z-z*v.y), (z*v.x-x*v.z), (x*v.y-y*v.x) );
        };

   
      // return magnitude

      inline float length( void )
        {
          return sqrt( (x*x) + (y*y) + (z*z) );
        };


      /*------------------------ VECTOR CONSTRUCTORS ------------------------*/

      inline cVector3f( void )
        {
          x=0; y=0; z=0; w=1;
        };

      inline cVector3f( float a , float b , float c )
        {
          x=a; y=b; z=c; w=1;
        };

      /*--------------------------- VECTOR MATHS ----------------------------*/


      // addition

      inline cVector3f operator+( const cVector3f &v ) const {
          return cVector3f( x+v.x , y+v.y , z+v.z );
        };

      inline cVector3f operator+( const float f ) const {
          return cVector3f ( x+f , y+f , z+f );
        };

      inline cVector3f& operator+=( const cVector3f &v )
        {
          x += v.x; y += v.y; z += v.z;
          return *this;
        };


      // subtration

      inline cVector3f operator-( const cVector3f &v ) const
        {
          return cVector3f( x-v.x , y-v.y , z-v.z );
        };

      inline cVector3f operator-( const float f ) const
        {
          return cVector3f( x-f , y-f , z-f );
        };

      inline cVector3f& operator-=( const cVector3f &v )
        {
          x -= v.x; y -= v.y; z -= v.z;
          return *this;
        };


      // reverse

      inline cVector3f operator-() const
        {
          return cVector3f( -x , -y , -z );
        };


      // multiplication

      inline cVector3f operator*( const float f ) const
        {
          return cVector3f( x*f , y*f , z*f );
        };

      inline cVector3f& operator*=( const float f )
        {
          x *= f; y *= f; z *= f;
          return *this;
        };


      // division

      inline cVector3f operator/( const float f ) const
        {
          float inv = 1.0f/f;
          return cVector3f(x*inv, y*inv, z*inv);
        };

      inline cVector3f& operator/=( const float f )
        {
          float invert = 1.0f / f;
          x *= invert; y *= invert; z *= invert;
          return *this;
        };


      // equality

      inline cVector3f& operator=( const cVector3f &v )
        {
          x = v.x; y = v.y; z = v.z;
          return *this;
        };

      inline cVector3f normalized( void )
        {
          return (*this / this->length());
        };

      inline void normalize( void ) 
        {
          float l = length();
          x=x/l; y=y/l; z=z/l;
        };


};


/*##############################################################################
###                            PLANE4F DATA CLASS
##############################################################################*/

/*
 *    **************************** Usage : *********************************
 *
 *       ------------ plane4f is based on plane equation : --------------
 *
 *     (plane4f.a)*x + (plane4f.b)*y + (plane4f.c)*z + (plane4f.d) = 0
 *
 *
 *       --------------------- 5 constructors exist ---------------------
 *
 *    plane4f p;
 *    plane4f p( float a, float b, float c, float d);
 *    plane4f p( cVector3f normal , float d );
 *    plane4f p( plane4f plane );
 *    plane4f p( cVector3f V1 , cVector3f V2 , cVector3f V3 );
 *
 *
 *       -------------------------- Operators ---------------------------
 *
 *    plane4f a,b;
 *    a = b;
 *
 *
 *       ----------------------- Member Functions -----------------------
 *
 *    plane4f a;
 *    cVector3f Normal;
 *
 *    Normal = a.normal();         : returns normal vector of plane
 *    a.setNormal( Normal );       : sets the normal vector of the plane
 *    a.distance( cVector3f Point ) : Performs a dot product with the normal vector
 *                                   and a point in space. It returns the distance
 *                                   from the plane. ( + or - float )
 *
 *    a.Valid();                   : returns a bool to see if plane equation is 
 *                                   valid or not
 *
 *    a.lineIntersection( cVector3f origin , cVector3f direction );
 *
 *                                 : returns the parametric t value for where the 
 *                                   plane interesects the line segment
 *
*/

class plane4f
  {

  public :

    float a;
    float b;
    float c;
    float d;

    /*---------------------------- CONSTRUCTORS ----------------------------*/

    inline plane4f( void ) {
      a = 0; b = 0; c = 0; d = 0;
    };

    inline plane4f( float A , float B , float C , float D ) {
      a = A; b = B; c = C; d = D;
    };

    inline plane4f( cVector3f v , float D ) {
      a = v.x; b = v.y; c = v.z; d = D;
    };

    inline plane4f( const plane4f &p ) {
      a = p.a; b = p.b; c = p.c; d = p.d;
    };

    plane4f( cVector3f v1 , cVector3f v2 , cVector3f v3 ) {
      cVector3f n = ( ( v3 - v2 ) ^ ( v1 - v2 ) ).normalized();
      setNormal(n);
      d = -(n*v1);
    };


    /*------------------------------ OPERATORS -----------------------------*/

    inline plane4f& operator=( const plane4f& p ) {
      a = p.a; b = p.b; c = p.c; d = p.d;
      return *this;
    };

    /*------------------------------ FUNCTIONS -----------------------------*/

    inline cVector3f normal( void ) {
      return cVector3f( a , b , c );
    };

    inline void setNormal( cVector3f n ) {
      a = n.x; b = n.y; c = n.z;
    };

    float distance( cVector3f point ) {
      return ( normal() * point ) + d;
    };

    float lineIntersection( cVector3f origin , cVector3f direction ) {
      return -((normal() * origin) + d)/(normal() * direction);
    };

    bool valid() {
      float l = normal().length();
      return ( ( l > 0.0f ) && ( ( 1.0f/l ) > 0.0f ) );
    };

};

/*##############################################################################
###                             matrix class
##############################################################################*/

class matrix

  {
    public :


      // data elements : prefixed with r for row

      float  r0_0,  r0_1,  r0_2,  r0_3;
      float  r1_0,  r1_1,  r1_2,  r1_3;
      float  r2_0,  r2_1,  r2_2,  r2_3;
      float  r3_0,  r3_1,  r3_2,  r3_3;



      // constructors

      inline matrix(void)
        {
          r0_0 =  1  ; r0_1 =  0  ; r0_2 =  0  ; r0_3 =  0  ;
          r1_0 =  0  ; r1_1 =  1  ; r1_2 =  0  ; r1_3 =  0  ;
          r2_0 =  0  ; r2_1 =  0  ; r2_2 =  1  ; r2_3 =  0  ;
          r3_0 =  0  ; r3_1 =  0  ; r3_2 =  0  ; r3_3 =  1  ;
        };

      inline matrix( const matrix &m )
        {
          r0_0 = m.r0_0;   r0_1 = m.r0_1;   r0_2 = m.r0_2;   r0_3 = m.r0_3;
          r1_0 = m.r1_0;   r1_1 = m.r1_1;   r1_2 = m.r1_2;   r1_3 = m.r1_3;
          r2_0 = m.r2_0;   r2_1 = m.r2_1;   r2_2 = m.r2_2;   r2_3 = m.r2_3;
          r3_0 = m.r3_0;   r3_1 = m.r3_1;   r3_2 = m.r3_2;   r3_3 = m.r3_3;
        };

      // Load in an identity matrix for a laugh...

      inline void LoadIdentity(void)
        {
          r0_0 =  1  ; r0_1 =  0  ; r0_2 =  0  ; r0_3 =  0  ;
          r1_0 =  0  ; r1_1 =  1  ; r1_2 =  0  ; r1_3 =  0  ;
          r2_0 =  0  ; r2_1 =  0  ; r2_2 =  1  ; r2_3 =  0  ;
          r3_0 =  0  ; r3_1 =  0  ; r3_2 =  0  ; r3_3 =  1  ;
        };



      // matrix transforms - rotatation, scaling & translation

      inline matrix& scale(float x, float y, float z)
        {
          matrix m;
          m.r0_0  =  x ;
          m.r1_1  =  y ;
          m.r2_2  =  z ;
          return (*this) *= m;
        };

      inline matrix& scale( cVector3f& v )
        {
          matrix m;
          m.r0_0  =  v.x ;
          m.r1_1  =  v.y ;
          m.r2_2  =  v.z ;
          return (*this) *= m;
        };

      inline matrix& translate( float x, float y, float z )
        {
          matrix m;
          m.r0_3 = x;
          m.r1_3 = y;
          m.r2_3 = z;
          return (*this) *= m;
        };

      inline matrix& translate( cVector3f& v )
        {
          matrix m;
          m.r0_3 = v.x;
          m.r1_3 = v.y;
          m.r2_3 = v.z;
          return (*this) *= m;
        };

      inline matrix& rotateX( double degrees )
        {
          double  sr = sin( (degrees*PI_OVER_180) );
          double  cr = cos( (degrees*PI_OVER_180) );
          matrix m;
          m.r1_1 =    cr; 
          m.r1_2 = -1*sr;
          m.r2_1 =    sr;
          m.r2_2 =    cr;
          return (*this) *= m;
        };

      inline matrix& rotateY( double degrees )
        {
          double  sr = sin(degrees*PI_OVER_180);
          double  cr = cos(degrees*PI_OVER_180);
          matrix m;
          m.r0_0 =    cr;
          m.r0_2 =    sr;
          m.r2_0 = (-1)*(sr);
          m.r2_2 =    cr;
          return (*this) *= m;
        };

      inline matrix& rotateZ( double degrees )
        {
          double  sr = sin(degrees*PI_OVER_180);
          double  cr = cos(degrees*PI_OVER_180);
          matrix m;
          m.r0_0 = cr;
          m.r0_1 = -1*sr;
          m.r1_0 = sr;
          m.r1_1 = cr;
          return (*this) *= m;
        };




      // overloaded operators

      inline matrix& operator=(const matrix &m)
        {
          r0_0 = m.r0_0;   r0_1 = m.r0_1;   r0_2 = m.r0_2;   r0_3 = m.r0_3;
          r1_0 = m.r1_0;   r1_1 = m.r1_1;   r1_2 = m.r1_2;   r1_3 = m.r1_3;
          r2_0 = m.r2_0;   r2_1 = m.r2_1;   r2_2 = m.r2_2;   r2_3 = m.r2_3;
          return *this;
        };

      inline matrix& operator*= (matrix& b)
        {
	     return *this = (*this) * b;
        };

      friend matrix    operator*(const matrix& a, const matrix& b );
      friend cVector3f operator*(const matrix& m, const cVector3f& v);



  };

extern matrix operator*( const matrix& a , const matrix& b );
extern cVector3f operator*(const matrix& m, const cVector3f& v);


#endif    /* FRUITYMATH_H */
